
module Bush =
struct
	type 'a t =
		| []
		| (::) of 'a * 'a t t
	let rec map : 'a 'b. ('a -> 'b) -> 'a t -> 'b t =
	   fun f -> function
	   | [] -> []
	   | h::t -> f h::map (map f) t
	let rec fold : 'a. ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b =
		fun f at z -> match at with
		| [] -> z
		| a::att -> f a (fold (fold f) att z)
	let size at =
		fold (fun _ n -> n + 1) at 0
end


module Lush =
struct
	type 'a t =
	   | Nil
	   | Cons of 'a * 'a list t
	let rec map : 'a 'b. ('a -> 'b) -> 'a t -> 'b t =
	   fun f -> function
	   | Nil -> Nil
	   | Cons(h,t) -> Cons(f h,map (List.map f) t)
	let rec fold : 'a. ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b =
		fun f at z -> match at with
		| Nil -> z
		| Cons(a,alt) -> f a (fold (List.fold_right f) alt z)
	let size at =
		fold (fun _ n -> n + 1) at 0
end


type 'a t =
   | Item of 'a
   | List of 'a list t
let rec map : 'a 'b. ('a -> 'b) -> 'a t -> 'b t =
   fun f -> function
   | Item x -> Item (f x)
   | List l -> List (map (List.map f) l)
let rec to_list : 'a. 'a t -> 'a list = function
   | Item x -> [x]
   | List l -> List.concat (to_list l)
let rec fold : 'a. ('a -> 'b -> 'b) -> ('a -> 'a t) -> 'a t -> 'b -> 'b =
	fun f g at z -> match at with
	| Item x ->  g x
	| List l -> f (fold (List.fold_right f) g l z)

