
type zero = O
type 'a succ = S of 'a

type _ complete =
	| Empty : zero complete
	| Fork : 'a complete * 'a complete -> 'a succ complete

let rec depth : type a. a complete -> int =
	function
	| Empty -> 0
	| Fork (l,r) -> 1 + depth r	
	
type _ fibonacci =
	| Zero : zero fibonacci
	| One : zero succ fibonacci
	| Add : 'a fibonacci * 'a succ fibonacci -> 'a succ succ fibonacci

let rec sum : type a. a fibonacci -> int =
	function
	| Zero | One -> 1
	| Add (l,r) -> sum l + sum r	

type 'a bush =
	| []
	| (::) of 'a * 'a bush bush

let example =
	[ 4;
	[ 8; [5]; [[3]] ];
	[ [7]; []; [[[7]]] ];
	[ [ []; [[0]] ] ] ]

let rec map : 'a 'b. ('a -> 'b) -> 'a bush -> 'b bush =
	fun f ab -> match ab with
	| [] -> []
	| a :: abb -> f a :: map (map f) abb

let rec fold : 'a. ('a -> 'b -> 'b) -> 'a bush -> 'b -> 'b =
	fun f ab z -> match ab with
	| [] -> z
	| a :: abb -> f a (fold (fold f) abb z)

let size ab =
	fold (fun _ n -> n + 1) ab 0 
		
type (_,_) index =
	| Zero : ('z,'z) index
	| Head : ('a,'z) index -> ('a bush,'z) index
	| Tail : ('a bush,'z) index -> ('a,'z) index
	
let rec member : type a z. (a,z) index -> a bush -> z =
	fun i ab -> match i,ab with
	| _, [] -> raise Not_found
	| Zero, a::abb -> a
	| Head j, a::abb -> member j a 
	| Tail j, a::abb -> member j abb

# member Zero example;;
- : int = 4
# member (Tail Zero) example;;
- : int bush = [ 8; [5]; [[3]] ]
# member (Tail (Head (Tail (Tail Zero)))) example;;
- : int bush bush = [[3]]
# member (Tail (Tail (Tail (Tail Zero)))) example;;
Exception: Not_found.

(* 'c is unbound *)
type ('a,'b) chain =
	| Link of 'a * 'b 
	| Append of ('a,'c) chain * ('c,'b) chain

(* 'c is unbound *)
type ('a,'b) chain =
	| Link of 'a * 'b 
	| Append of 'p * 'q
constraint 'p = ('a,'c) chain
constraint 'q = ('c,'b) chain

(* 'c is existentially quantified *)
type ('a,'b) chain =
	| Link : 'a * 'b -> ('a,'b) chain 
	| Append : ('a,'c) chain * ('c,'b) chain -> ('a,'b) chain

let rec length : type a b. (a,b) chain -> int =
	function
	| Link _ -> 1
	| Append(p,q) -> length p + length q

